package org.ids_mannheim;

import java.io.BufferedReader;
import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import static java.lang.Thread.sleep;

public class Worker implements Runnable {
    private static final Pattern new_text_pattern = Pattern.compile("^#\\s+text_id\\s*=\\s*(.+)");
    private static final int MAX_RETRIES = 10;
    private final ArrayList<String> fnames;
    private final BlockingQueue<Integer> queue;
    private final ConcurrentMap<String, FoldedEntry> map;
    private final int folds;
    private final Progressbar etaPrinter;
    private final int ngram_size;
    private final Logger logger;
    private final WorkerNodePool pool;

    public Worker(BlockingQueue<Integer> queue, ArrayList<String> fnames, int ngram_size, int folds,
                  ConcurrentMap<String, FoldedEntry> map,
                  WorkerNodePool pool,
                  Progressbar etaPrinter, Logger logger) {
        this.queue = queue;
        this.fnames = fnames;
        this.map = map;
        this.ngram_size = ngram_size;
        this.folds = folds;
        this.pool = pool;
        this.etaPrinter = etaPrinter;
        this.logger = logger;
    }

    @Override
    public void run() {
        try {
            int index = queue.take();
            int retries = MAX_RETRIES;
            SlidingWindowQueue slidingWindowQueue = new SlidingWindowQueue(ngram_size, (s, f) -> FoldedEntry.incr(map, s, f));
            while (index >= 0) {
                String fname = fnames.get(index);
                File current_file = new File(fname);
                long file_size = current_file.length();
                int poolIndex = pool.getNextFree();
                logger.info(String.format("Started %d/%d %s %s", index, fnames.size(), pool.getHost(poolIndex), current_file.getName()));
                String[] cmd = {
                        "/bin/sh",
                        "-c",
                        pool.getExec(poolIndex) + "/usr/local/kl/bin/korapxml2conllu " + fname
                };
                Process p = Runtime.getRuntime().exec(cmd);
                BufferedReader in = new BufferedReader(new InputStreamReader(p.getInputStream()));
                String line;
                int fold = -1;
                int texts = 0;
                while ((line = in.readLine()) != null) {
                    if (line.startsWith("#")) {
                        Matcher matcher = new_text_pattern.matcher(line);
                        if (matcher.find()) {
                            fold = Math.abs(matcher.group(1).hashCode()) % folds + 1;
                        }
                        slidingWindowQueue.reset(fold);
                        texts++;
                        continue;
                    }
                    String[] strings = line.split("\\s+");
                    if (strings.length < 4) {
                        continue;
                    }
                    slidingWindowQueue.add(strings[1]);
                }
                pool.markFree(poolIndex);
                if (texts > 0) {
                    logger.info(pool.getHost(poolIndex) + " finished " + fname + " with " + texts + " texts");
                    etaPrinter.update(file_size);
                    retries = MAX_RETRIES;
                    index = queue.take();
                } else {
                    if (--retries > 0) {
                        logger.warning("Retrying " + fname);
                        sleep(1000);
                    } else {
                        logger.severe("Giving up " + fname);
                        index = queue.take();
                    }
                }
            }
        } catch (InterruptedException | IOException e) {
            Thread.currentThread().interrupt();
        }
    }
}
